/*
Copyright (C) 2004  The PARI group.

This file is part of the PARI package.

PARI/GP is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation. It is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY WHATSOEVER.

Check the License for details. You should have received a copy of it, along
with the package; see the file 'COPYING'. If not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#include "pari.h"
#include "paripriv.h"
#include "tree.h"
#include "opcode.h"
#include "anal.h"

entree * getentry(int n, long func);
void compilenode(int n, int mode);
GEN listtogen(int n, int f);

void
compilecast(int type, int mode)
{
  if (type==mode) return;
  if (type==Gsmall && mode==Ggen) op_push(OCstoi,0);
  if (type==Ggen && mode==Gsmall) op_push(OCitos,0);
}

void
compilelvalue(int n)
{ 
  if (tree[n].f==Fentry)
    return;
  else
  {
    int y=tree[n].y;
    int yx=tree[y].x;
    int yy=tree[y].y;
    long f=tree[y].f;
    compilenode(yx,Gsmall);
    if (f==Fmatrix && yy==-1)
      op_push(OCcompo1ptr,0);
    else
    {
      switch(f)
      {
      case Fmatrix:
        compilenode(yy,Gsmall);
        op_push(OCcompo2ptr,0);
        break;
      case FmatrixR:
        op_push(OCcompoCptr,0);
        break;
      case FmatrixL:
        pari_err(impl,"[_,]");
       // op_push(OCcompoLptr,0);
        break;
      }
    }
  } 
}

void 
compilefacteurmat(int n, int mode)
{ 
  int x=tree[n].x;
  int y=tree[n].y;
  int yx=tree[y].x;
  int yy=tree[y].y;
  long f=tree[y].f;
  compilenode(x,Ggen);
  if (f==Fmatrix && yy==-1)
  {
    compilenode(yx,Gsmall);
    op_push(OCcompo1,mode);
    return;
  }
  switch(f)
  {
  case Fmatrix:
    compilenode(yx,Gsmall);
    compilenode(yy,Gsmall);
    op_push(OCcompo2,mode);
    break;
  case FmatrixR:
    compilenode(yx,Gsmall);
    op_push(OCcompoC,mode);
    return;
  case FmatrixL:
    pari_err(impl,"[_,]");
    compilenode(yy,Gsmall);
    op_push(OCcompo1,mode);
    return;
  default:
    pari_err(bugparier,"compilefacteurmat");
  }
}

static GEN
cattovec(int n, int fnum)
{
  int x=n, i=0, nb;
  GEN stack;
  if (n==GNOARG) return cgetg(1,t_VECSMALL);
  while(1)
  {
    int xx=tree[x].x;
    int xy=tree[x].y;
    if (tree[x].f!=Fentryfunc || xx!=fnum) break;
    x=tree[xy].x;
    i++;
  }
  nb=i+1;
  stack=cgetg(nb+1,t_VECSMALL);
  for(x=n;i>0;i--)
  {
    int y=tree[x].y;
    x=tree[y].x;
    stack[i+1]=tree[y].y;
  }
  stack[1]=x;
  return stack;
}

static entree *
getentryfunc(int n)
{
  if (tree[n].f!=Fentryfunc) 
    pari_err(talker,"function required (found [%s])",Ffuncname[tree[n].f]);
  else
  {
    int x=tree[n].x;
    char *name=(char *)value[x].val.str;
    entree *ep; 
    ep = is_entry(name);
    if (!ep)
      ep=fetch_named_var(name);
    return ep;
  }
  return NULL;/*NOT REACHED*/
}

static void 
compilefunc(int n, int mode)
{
  long i,j;
  int x=tree[n].x;
  char *name=(char *)value[x].val.str;
  int ret;
  char const *p,*q;
  char c;
  PPproto mod;
  GEN arg=listtogen(tree[n].y,Flistarg);
  long nbpointers=0;
  long nb=lg(arg)-1;
  entree *ep = is_entry(name);
  if (!ep) 
    pari_err(talker,"unknown function %s",name);
  ep=do_alias(ep);
  if (EpVALENCE(ep)==EpVAR || EpVALENCE(ep)==EpGVAR)
    pari_err(talker,"%s is not a function in function call", name);
  if (ep->valence==EpUSER)
  {
    for (j=1;j<=nb;j++)
      if (arg[j]!=GNOARG)
        compilenode(arg[j], Ggen);
      else
        op_push(OCpushstoi,0);
    for (   ;j<=ep->arity;j++)
      op_push(OCpushstoi,0);
    op_push(OCcalluser, (long) ep);
    compilecast(Ggen,mode);
    return;
  }
  if (!strcmp(name,"O"))
  {
    int a=tree[n].y;
    if (tree[a].f!=Fentryfunc || tree[a].x!=OPpow)
    {
      compilenode(a,Ggen);
      op_push(OCpushlong,1);
    }
    else
    {
      int y=tree[a].y;
      compilenode(tree[y].x,Ggen);
      compilenode(tree[y].y,Gsmall);
    }
    op_push(OCpushlong,2);
    op_push(OCcallgen,(long)ep);
    compilecast(Ggen,mode);
    return;
  }
  p=ep->code;
  if (!ep->value)
    pari_err(impl,name);
  if      (*p < 'a') { ret = RET_GEN; }
  else if (*p == 'v') { ret = RET_VOID; p++; }
  else if (*p == 'i') { ret = RET_INT;  p++; }
  else if (*p == 'l') { ret = RET_LONG; p++; }
  else ret = RET_GEN;
  i=0; j=1;
  while(*p=='G' && j<=nb)
  {
    compilenode(arg[j++], Ggen);
    p++; i++;
  }
  if (*p)
  {
    q=p; 
    while((mod=parseproto(&p,&c))!=PPend)
    {
      if (j<=nb && arg[j]!=GNOARG && (mod==PPdefault || mod==PPdefaultmulti))
        mod=PPstd;
      switch(mod)
      {
      case PPstd:
        if (j>nb)
          pari_err(talker,"missing mandatory argument for %s",name);
        switch(c)
        {
        case 'G':
          compilenode(arg[j++],Ggen);
          break;
        case 'M':
        case 'L':
          compilenode(arg[j++],Gsmall);
          break;
        case 'n':
          compilenode(arg[j++],Ggen);
          op_push(OCvarn, 0);
          break;
        case '&':
          {
            entree *ep;
            int a=arg[j++],b;
            if (tree[a].f==Frefarg)
              a=tree[a].x;
            b=a;
            while(tree[a].f!=Fentry) a=tree[a].x;
            ep=getentry(a,0);
            op_push(OCnewptr, (long) ep);
            compilelvalue(b);
            op_push(OCpushptr, 0);
            nbpointers++;
            break;
          }
          break;
        case 'I':
        case 'E':
          {
            long sop,sdat,off;
            getcodepos(&sop,&sdat,&off);
            compilenode(arg[j++],Ggen);
            op_push(OCpushgen, data_push(getclosure(sop,sdat,off)));
            break;
          }
        case 'V':
        case 'S':
          {
            entree *ep = getentry(arg[j++],c=='S');
            op_push(OCpushlong, (long)ep);
            break;
          }
        case 'r':
          {
            int a=arg[j++];
            if (tree[a].f==Fentry)
            {
              op_push(OCpushgen,data_push(strtoGENstr(value[tree[a].x].val.str)));
              op_push(OCtostr, 0);
            }
            else
            {
              compilenode(a,Ggen);
              op_push(OCtostr, 0);
            }
            break;
          }
        case 's':
          compilenode(arg[j++],Ggen);
          op_push(OCtostr, 0);
          break;
          break;
        default:
          pari_err(talker,"Unknown prototype code `%c' for `%s'",c,name);
        }
        break;
      case PPauto:
        switch(c)
        {
        case 'p':
          op_push(OCprecreal,0);
          break;
        case 'P':
          op_push(OCprecdl,0);
          break;
        }
        break;
      case PPdefault:
        j++;
        switch(c)
        {
        case 'G':
        case '&':
        case 'V':
        case 'r':
        case 'E':
        case 'I':
          op_push(OCpushlong,0);
          break;
        case 'n':
          op_push(OCpushlong,-1);
          break;
        default:
          pari_err(talker,"Unknown default in prototype code `%c' for `%s'",c,name);
        }
        break;
      case PPdefaultmulti:
        j++;
        switch(c)
        {
        case 'L':
        case 'M':
          op_push(OCpushlong,strtol(q+1,NULL,10));
          break;
        case 'r':
        case 's':
          op_push(OCpushlong,(long)"");
          break;
        default:
          pari_err(talker,"Unknown default in prototype code `%c' for `%s'",c,name);
        }
        break;
      case PPstar:
        switch(c)
        {
        case 'V':
          {
            int x=tree[arg[j]].x;
            int y=tree[arg[j]].y;
            entree *ep;
            if (tree[arg[j]].f!=Faffect)
              pari_err(talker,"= expected");
            ep = getentry(x,0);
            op_push(OCpushlong, (long)ep);
            compilenode(y,Ggen);
            i++; j++;
          }
          break;
        case 'I':
        case 's':
          {
            long n=nb+1-j;
            long k,l,l1,m;
            GEN g=cgetg(n+1,t_VEC);
            for(l1=0,k=1;k<=n;k++)
            {
              gel(g,k)=cattovec(arg[j+k-1],OPcat);
              l1+=lg(g[k])-1;
            }
            for(m=1,k=1;k<=n;k++)
            {
              for(l=1;l<lg(g[k]);l++)
                compilenode(mael(g,k,l),Ggen);
            }
            j=nb+1;
            op_push(OCvec, l1);
            break;
          }
        default:
          pari_err(talker,"Unknown prototype code `%c*' for `%s'",c,name);
        }
        break;
      default:
        pari_err(bugparier,"PPproto %d in gencallfunc",mod);
      }
      i++;
      q=p;
    }
  }
  if (j<=nb)
    pari_err(talker,"too many arguments for %s",name);
  op_push(OCpushlong, i);
  switch (ret)
  {
  case RET_GEN: 
    op_push(OCcallgen, (long) ep);
    compilecast(Ggen,mode);
    break;
  case RET_INT:
    op_push(OCcallint, (long) ep);
    compilecast(Gsmall,mode);
    break;
  case RET_LONG:
    op_push(OCcalllong, (long) ep);
    compilecast(Gsmall,mode);
    break;
  case RET_VOID:
    op_push(OCcallvoid, (long) ep);
    compilecast(Ggen,mode);
    break;
  }
  op_push(OCendptr,nbpointers);
}

static void
compilevec(int n)
{
  pari_sp ltop=avma;
  int x=tree[n].x;
  long i;
  GEN arg=listtogen(x,Fmatrixelts);
  long l=lg(arg)-1;
  for(i=1;i<=l;i++)
    compilenode(arg[i],Ggen);
  avma=ltop;
  op_push(OCvec,l);
}

static void
compilemat(int n)
{
  pari_sp ltop=avma;
  int x=tree[n].x;
  long i,j;
  GEN line=listtogen(x,Fmatrixlines);
  long nbline = lg(line)-1, nbcol=0;
  GEN lines=cgetg(nbline+1,t_MAT);
  for(i=1;i<=nbline;i++)
  {
    long k;
    lines[i]=(long)listtogen(line[i],Fmatrixelts);
    k=lg(lines[i])-1;
    if (i>1 && k!=nbcol)
      pari_err(talker,"Matrix must be rectangular");
    nbcol=k;
  }
  for(i=1;i<=nbcol;i++)
  {
    for(j=1; j<=nbline; j++)
      compilenode(mael(lines,j,i), Ggen);
    op_push(OCcol,nbline);
  }
  op_push(OCmat,nbcol);
  avma=ltop;
}

void
compilenode(int n, int mode)
{
  int x,y;
  if (n<=0)
    pari_err(talker,"missing mandatory argument");
  x=tree[n].x;
  y=tree[n].y;
  if (DEBUGLEVEL>4)
    fprintferr("[%d]:%s(%d,%d)",n,Ffuncname[tree[n].f],x,y);
   
  switch(tree[n].f)
  {
  case Fseq:
    compilenode(x,Gvoid);
    op_push(OCpop,1);
    compilenode(y,mode);
    return;
  case Fmatrix:
  case FmatrixL:
  case FmatrixR:
    break;
  case Ffacteurmat:
    compilefacteurmat(n,mode);
    break;
  case Faffect:
    compilefunc(newnode(Fentryfunc,newentry("_=_"),newnode(Flistarg,x,y)),mode);
    break;
  case Fconst:
    if (mode==Gsmall)
      pari_err(talker,"small integer expected");
    switch(value[x].type)
    {
    case CSTreal: 
      op_push(OCpushreal, data_push(value[x].val.gen));
      return;
    case CSTgen: 
      op_push(OCpushgen, data_push(value[x].val.gen));
      return;
    case CSTstr:
      op_push(OCpushgen, data_push(strtoGENstr((char *)value[x].val.str)));
      return;
    case CSTquote:
      op_push(OCpushgen, data_push(pol_x(fetch_user_var((char *)value[x].val.str))));
      return;
    }
    break;
  case Fsmall:
    if (mode==Gsmall)
      op_push(OCpushlong, x);
    else
      op_push(OCpushstoi, x);
    return;
  case Fmatrixelts:
  case Fmatrixlines:
    break;
  case Fvec:
    if (mode==Gsmall)
      pari_err(talker,"small integer expected");
    compilevec(n);
    return;
  case Fmat:
    if (mode==Gsmall)
      pari_err(talker,"small integer expected");
    compilemat(n);
    return;
  case Flistarg:
    break;
  case Frefarg:
    pari_err(talker,"unexpected &"); 
    break;
  case Fentry:
    {
      char *name=(char *)value[x].val.str;
      entree *ep=is_entry(name);
      if (ep && EpVALENCE(ep)!=EpVAR && EpVALENCE(ep)!=EpGVAR)
        compilefunc(n,mode);
      else
      {
        op_push(OCpushentryval,(long)getentry(n,0));
        compilecast(Ggen,mode);
      }
      return;
    }
  case Fentryfunc:
    compilefunc(n, mode);
    return;
  case Fdeffunc:
    {
      long sop,sdat,off;
      long i;
      GEN arg2=listtogen(tree[x].y,Flistarg);
      entree *ep=getentryfunc(x);
      /* Check for local() declarations.*/
      int loc=y;
      GEN vars=NULL;
      while (tree[loc].f==Fseq) loc=tree[loc].x;
      ep->valence=EpUSER;
      ep->arity=lg(arg2)-1;
      getcodepos(&sop,&sdat,&off);
      for (i=1;i<lg(arg2);i++)
      {
        long a = arg2[lg(arg2)-i];
        if (tree[a].f==Faffect)
          a=tree[a].x;
        op_push(OCpushargs,(long)getentry(a,0));
      }
      if (tree[loc].f==Fentryfunc 
          && strcmp(value[tree[loc].x].val.str,"local")==0)
      {
        vars=listtogen(tree[loc].y,Flistarg);
        for (i=1;i<lg(vars);i++)
        {
          int a=vars[i];
          if (tree[a].f==Faffect)
          {
            compilenode(tree[a].y,Ggen);
            a=tree[a].x;
          }
          else
            op_push(OCpushstoi,0);
          op_push(OCpushargs,(long)getentry(a,0));
        }
      }
      compilenode(y,Ggen);
      for (i=1;i<lg(arg2);i++)
      {
        long a = arg2[i];
        if (tree[a].f==Faffect)
          a=tree[a].x;
        op_push(OCpopargs,(long)getentry(a,0));
      }
      if (vars)
      {
        for (i=1;i<lg(vars);i++)
        {
          int a=vars[i];
          if (tree[a].f==Faffect)
            a=tree[a].x;
          op_push(OCpopargs,(long)getentry(a,0));
        }
      }
      op_push(OCpushgen, data_push(getclosure(sop,sdat,off)));
      op_push(OCdeffunc, (long) ep);
      break;
    }
  case Ftag:
    compilenode(x, mode);
    return;
  default:
    pari_err(talker,"node %s not implemented yet\n",Ffuncname[tree[n].f]);
  }
}

GEN gp_closure(int n)
{
  compilenode(n,Ggen);
  return getclosure(0,0,-1); 
}

static long
check_array_index(long c, long max)
{
  if (c < 1 || c >= max)
  {
    char s[80];
    sprintf(s,"array index (%ld) out of allowed range ",c);
    if (max == 1) strcat(s, "[none]");
    else if (max == 2) strcat(s, "[1]");
    else sprintf(s,"%s[1-%ld]",s,max-1);
    pari_err(talker,s);
  }
  return c;
}

INLINE void
copyvalue(entree *ep, GEN x) {
  new_val_cell(ep, x, typ(x) >= t_VEC ? COPY_VAL: PUSH_VAL);
}
INLINE void
pushvalue(entree *ep, GEN x) { new_val_cell(ep, x, PUSH_VAL); }

#define killvalue(v) pop_val(get_ep(v))

static long st[1024];
static long sp=0;

GEN
closure_eval(GEN C)
{
  pari_sp ltop=avma;
  gp_pointer ptrs[8];
  long rp=0;
  GEN code=gel(C,1);
  GEN data=gel(C,2);
  long i,j;
  for(i=1;i<lg(code);i+=2)
  {
    long opcode=code[i];
    long operand=code[i+1];
    entree *ep;
    switch(opcode)
    {
    case OCpushlong:
        st[sp++]=operand;
        break;
    case OCpushgen:
        gel(st,sp++)=gel(data,operand);
        break;
    case OCpushreal:
        gel(st,sp++)=strtor(GSTR(data[operand]),precreal);
        break;
    case OCpushstoi:
        gel(st,sp++)=stoi(operand);
        break;
    case OCpushentryval:
        ep=(entree*)operand;
        gel(st,sp++)=ep->value;
        break;
    case OCnewptr:
        {
          gp_pointer *g = &ptrs[rp++];
          g->ep = (entree*) operand;
          g->x = (GEN) g->ep->value;
          break;
        }
    case OCpushptr:
        {
          gp_pointer *g = &ptrs[rp-1];
          gel(st,sp++) = (GEN)&(g->x); 
        }
        break;
    case OCendptr:
        for(j=0;j<operand;j++)
        {
          gp_pointer *g = &(ptrs[--rp]);
          if (g->ep) changevalue(g->ep, g->x);
          else (void)change_compo(avma, &(g->c), g->x);
        }
        break;
    case OCprecreal:
        st[sp++]=precreal;
        break;
    case OCprecdl:
        st[sp++]=precdl;
        break;
    case OCstoi:
        gel(st,sp-1)=stoi(st[sp-1]);
        break;
    case OCitos:
        st[sp-1]=itos(gel(st,sp-1));
        break;
    case OCtostr:
        st[sp-1]=(long) GSTR(GENtoGENstr(gel(st,sp-1)));
        break;
    case OCvarn:
        st[sp-1]=gvar(gel(st,sp-1));
        break;
    case OCcompo1:
        {
          GEN  p=gel(st,sp-2);
          long c=st[sp-1];
          sp--;
          switch(typ(p))
          {
          case t_VEC: case t_COL:
            check_array_index(c, lg(p));
            gel(st,sp-1) = gel(p,c);
            break;
          case t_LIST:
            check_array_index(c,lgeflist(p)-1);
            gel(st,sp-1) = gel(p,c+1);
            break;
          case t_VECSMALL:
            check_array_index(c,lg(p));
            gel(st,sp-1) = stoi(p[c]);
            break;
          case t_MAT:
            pari_err(talker,"matrix has no simple components");
            break;
          default:
            pari_err(talker,"object has no components");
            break;
          }
          break;
        }
    case OCcompo1ptr:
        { 
          long c=st[sp-1];
          gp_pointer *g = &ptrs[rp-1];
          matcomp *C=&g->c;
          GEN p;
          if (g->ep)
          {
            g->ep=NULL;
            C->full_col = C->full_row = 0;
            C->parent   = (GEN)    g->x;
            C->ptcell   = (GEN *) &g->x;
          }
          p=*C->ptcell;
          sp--;
          switch(typ(p))
          {
          case t_VEC: case t_COL: case t_VECSMALL:
            check_array_index(c, lg(p));
            C->ptcell = (GEN *) p+c;
            break;
          case t_LIST:
            check_array_index(c,lgeflist(p)-1);
            C->ptcell = (GEN *) p+c+1;
            break;
          case t_MAT: 
            pari_err(talker,"matrix has no simple components");
          default:
            pari_err(talker,"object has no components");
          }
          g->x = *(g->c.ptcell);
          break;
        }
    case OCcompo2:
        {
          GEN  p=gel(st,sp-3);
          long c=st[sp-2];
          long d=st[sp-1];
          if (typ(p)!=t_MAT)
            pari_err(talker,"[_,_]: not a matrix");
          check_array_index(d, lg(p));
          check_array_index(c, lg(p[d]));
          sp-=2;
          gel(st,sp-1) = gcopy(gcoeff(p,c,d));
          break;
        }
    case OCcompo2ptr:
        {
          long b=st[sp-2];
          long c=st[sp-1];
          gp_pointer *g = &ptrs[rp-1];
          matcomp *C=&g->c;
          GEN p;
          if (g->ep)
          {
            g->ep=NULL;
            C->full_col = C->full_row = 0;
            C->parent   = (GEN)    g->x;
            C->ptcell   = (GEN *) &g->x;
          }
          p=*C->ptcell;
          sp-=2;
          if (typ(p)!=t_MAT)
            pari_err(talker,"[_,_]: not a matrix");
          check_array_index(b, lg(p));
          check_array_index(c, lg(p[b]));
          C->ptcell = (GEN *) gel(p,c)+b;
          g->x = *(g->c.ptcell);
          break;
        }
    case OCcompoC:
        {
          GEN  p=gel(st,sp-2);
          long c=st[sp-1];
          if (typ(p)!=t_MAT)
            pari_err(talker,"[_,]: not a matrix");
          check_array_index(c, lg(p));
          sp--;
          gel(st,sp-1) = gcopy(gel(p,c));
          break;
        }
    case OCcompoCptr:
        {
          long c=st[sp-1];
          gp_pointer *g = &ptrs[rp-1];
          matcomp *C=&g->c;
          GEN p;
          if (g->ep)
          {
            g->ep=NULL;
            C->full_col = C->full_row = 0;
            C->parent   = (GEN)    g->x;
            C->ptcell   = (GEN *) &g->x;
          }
          p=*C->ptcell;
          sp--;
          if (typ(p)!=t_MAT)
            pari_err(talker,"[_,]: not a matrix");
          check_array_index(c, lg(p));
          C->ptcell = (GEN *) p+c;
          g->x = *(g->c.ptcell);
          break;
        }

    case OCpushargs:
        ep=(entree *)operand;
        copyvalue(ep,gel(st,sp-1));
        sp--;
        break;
    case OCpopargs:
        ep=(entree *)operand;
        pop_val(ep);
        break;
#define ARGS st[sp-n-1],st[sp-n],st[sp-n+1],st[sp-n+2],st[sp-n+3],st[sp-n+4],st[sp-n+5],st[sp-n+6]
    case OCcallgen:
        {
          long n=st[sp-1];
          entree *ep=(entree*)operand;
          void *call=ep->value;
          gp_function_name=ep->name;
          GEN res = ((GEN (*)(ANYARG))call)(ARGS);
          if (!res) return NULL;
          gp_function_name=NULL;
          sp-=n+1;
          gel(st,sp++)=res;
          break;
        }
    case OCcalllong:
        {
          long n=st[sp-1];
          entree *ep=(entree*)operand;
          void *call=ep->value;
          gp_function_name=ep->name;
          long res = ((long (*)(ANYARG))call)(ARGS);
          if (br_status!=br_NONE) return NULL;
          gp_function_name=NULL;
          sp-=n+1;
          st[sp++] = res;
          break;
        }
    case OCcallint:
        {
          long n=st[sp-1];
          entree *ep=(entree*)operand;
          void *call=ep->value;
          gp_function_name=ep->name;
          long res = ((int (*)(ANYARG))call)(ARGS);
          if (br_status!=br_NONE) return NULL;
          gp_function_name=NULL;
          sp-=n+1;
          st[sp++] = res;
          break;
        }
    case OCcallvoid:
        {
          long n=st[sp-1];
          entree *ep=(entree*)operand;
          void *call=ep->value;
          gp_function_name=ep->name;
          ((void (*)(ANYARG))call)(ARGS);
          if (br_status!=br_NONE) return NULL;
          gp_function_name=NULL;
          sp-=n+1;
          gel(st, sp++) = gnil;
          break;
        }
    case OCcalluser:
        {
          GEN z;
          ep = (entree*) operand;
          gp_function_name=ep->name;
          z = closure_eval((GEN) ep->value);
          if (!z) 
          {
            if (br_status==br_RETURN)
            {
              br_status=br_NONE;
              z=gcopy(br_res);
            }
            else pari_err(talker, "break not allowed here");
          }
          gel(st, sp++) = z;
          gp_function_name=NULL;
          break;
        }
    case OCvec:
    case OCcol:
    case OCmat:
        {
          long j;
          GEN v=cgetg(operand+1,(opcode==OCvec?t_VEC:opcode==OCcol?t_COL:t_MAT));
          for (j=1;j<=operand;j++)
            gel(v,j)=gel(st,sp-operand+j-1);
          sp-=operand;
          gel(st,sp++)=v;
          break;
        }
    case OCdeffunc:
        ep=(entree*)operand;
        ep->value = (void *) gclone(gel(st,sp-1));
        gel(st, sp-1) = gnil;
        break;
    case OCpop:
        sp-=operand;
        break;
    default:
        pari_err(bugparier,"closure_eval");
    }
  }
  sp--;
  if (sp<0) pari_err(bugparier,"closure_eval");
  if (gel(st,sp)==gnil) {avma=ltop; return gnil;}
  return gerepilecopy(ltop,gel(st,sp)); 
}

GEN
closure_evalnobrk(GEN C)
{
  return closure_eval(C);
}

GEN
closure(GEN z)
{
  return z;
}
